module ComputationM where

import Types
import Sources
import Functions

import Unsafe.Coerce (unsafeCoerce)
import Control.Monad.State
import Control.Monad.Writer

data TypedRun a b where
  Function :: IsFunction f a b => f -> TypedRun a b
  FunctionIO :: IsFunctionIO f a b => f -> TypedRun a b
  Inject :: b -> TypedRun () b
  GetListElem :: TypedRun [b] b
  SetListElem :: TypedRun a ()
  MapComp :: [DependencyUntyped] -> TypedRun [a] [b]

deriving instance (Show a, Show b) => Show (TypedRun a b)

data Token a where
  Token :: Int -> Token a
  NoToken :: Token ()

deriving instance Show (Token a)

data Dependency a b = Dependency (Token a) (TypedRun a b) (Token b)
  deriving (Show)

data DependencyUntyped where
  DependencyUntyped :: Dependency a b -> String -> DependencyUntyped

instance Show DependencyUntyped where
  show (DependencyUntyped _ s) = s

newtype ComputationM' a = ComputationM { unComputationM :: WriterT [DependencyUntyped] (State Int) a }
  deriving (Functor, Applicative, Monad, MonadState Int, MonadWriter [DependencyUntyped])

type ComputationM a = ComputationM' (Token a)

evalComputationM' :: Int -> ComputationM () -> ([DependencyUntyped], Int)
evalComputationM' top m = runState (execWriterT (unComputationM m)) top

evalComputationM :: ComputationM () -> [DependencyUntyped]
evalComputationM m = fst (evalComputationM' 0 m)

genDependency :: (Show u, Show a) => (Token a -> Dependency u a) -> ComputationM a
genDependency g = do
  top <- get
  let top' = top + 1
      target = Token top'
  put top'
  let result = g target
  tell [DependencyUntyped result (show result)]
  pure target

inject :: Show a => a -> ComputationM a
inject x = genDependency (Dependency NoToken (Inject x))

getListElem :: Show a => Token [a] -> ComputationM a
getListElem outer = genDependency (Dependency outer GetListElem)

setListElem :: Show a => Token a -> ComputationM ()
setListElem a = genDependency (Dependency a SetListElem)

run :: (Show a, Show b, IsFunction f a b) => f -> Token a -> ComputationM b
run f input = genDependency (Dependency input (Function f))

isImageFilename = run IsImageFilename
convertedImageFilename = run ConvertedImageFilename

runIO :: (Show a, Show b, IsFunctionIO f a b) => f -> Token a -> ComputationM b
runIO f input = genDependency (Dependency input (FunctionIO f))

listDirectory = runIO ListDirectory
openImage = runIO OpenImage
convertImage = runIO ConvertImage
saveFile = runIO SaveFile
runPandoc = runIO RunPandoc

mapComputationM :: (Show a, Show b) => (Token a -> ComputationM b) -> Token [a] -> ComputationM [b]
mapComputationM f input = do
  top <- get
  let (res, top') = evalComputationM' top m
  put top'
  genDependency (Dependency input (MapComp res))
  where m :: ComputationM ()
        m = do
          inp <- getListElem input
          outp <- f inp
          setListElem outp
